/* Temerature control using a Fuzzy Logic Controller
 *  
 * read temperature using 
 * DS18B20 digital sensor
 * 
 * Shield LCD 16X2
 * GBR / 03.12.2016
 */ 

# include <OneWire.h> // for one-wire devices - temperature sensor
# include <DallasTemperature.h> // for temperature sensor
# include <LiquidCrystal.h> // for 16 x 2 LCD display
# include <FuzzyGbr.h> // for fls, membership functions
# include <Adafruit_MCP4725.h> // for DAC convertor, used to output the control voltage


# define tempData 8 // digital pin for data, Arduino Mega/Uno
// initialize an instance of OneWire library in our code
OneWire oneWire(tempData);
// Pass our oneWire reference to Dallas Temperature library.
DallasTemperature sensors(&oneWire);

// initialize an instance of LiquidCristal library in our code
  LiquidCrystal LCDdisplay(7,6,5,4,3,2);

  //initialize an instance of the class with membership functions
FuzzyGbr FuzzyMf;

Adafruit_MCP4725 dacI2C;  // constructor; an instance of the library

// ----------------------------------------------------------------
float tempSetPoint=30; // Set point temperature
float tempRead;  // temperature read from sensor
// set the gain for error, chang-of-error and du for fls

// crisp inputs for fuzzy sets
float err; // current crisp input value for error
float errFls; //current crisp error for fls input
float errPrev = 0; // previous value for error
float cerr; // current value for change-of-error
float cerrFls; //current crisp change-of-error for fls input
float du;         // current crisp output of control fls
float u = 4095;   // curent output (control signal); starts from 4095 - no power into the load

float gainErr = 0.15;
float gainCerr = 1.5;
float gainDu = 1000;

// mf for error stored in a 2D array, trimf, 3 params in a row
float errMf[3][3] = {  { -1, -1, 0},
                       { -1,  0, 1},
                       { 0,  1,  1}   };

// mf for chang-of-error stored in a 2D array, trimf, 3 params in a row
float cerrMf[3][3] = { { -1, -1, 0},
                       { -1,  0, 1},
                       { 0,  1,  1}   };

// output singleton mf stored in a 2D array, similar with the fuzzy rules table 
float duMf[3][3] = { { -1, -1, 0},
                     { -1,  0, 1},
                     {  0,  1, 1}   };


 // intermediate variable declarations
float mdMfErr[3]; // membership degree for current error to its fuzzy sets
float mdMfCerr[3]; // membership degree for current change-of-error to its fuzzy sets
float fdRule[3][3]; // firing degree for rules
float ws; // weighted sum for defuzzification
float sfd; // sum of firing degrees for deffuzification
//float du; // change-of-control (delta) (output of the fuzzy system)
int i; // index
int j; // index        
float controlVoltage;  // echivalent control voltage generated by the DAC     
char sign[3]; // to display on LCD <, >, etc.
    

void setup() {
  // put your setup code here, to run once:
  Serial.begin(9600);  // start serial port
  // Serial.println("Temperature measurement with DS18B20");
  sign [2] = NULL;
   
  sensors.begin();  // Start up the DS18B20 library
  LCDdisplay.begin(16,2);  // Start up the LCD shield library
  dacI2C.begin(0x62); // start dac @ 0x62 hex I2C address
  
  LCDdisplay.setCursor(0,0); // initial position for text, 1nd row
  LCDdisplay.print("T_Ref");
  LCDdisplay.setCursor(0,1); // initial position for text, 2nd row
  LCDdisplay.print(tempSetPoint);
  LCDdisplay.setCursor(5,1); // initial position for text, 2nd row
  LCDdisplay.print(" C");

  Serial.println("SetPoint   Temp    err    cerr      errFls     cerrFls         du         u     controlVoltage");
  }

// function for multiplication with gain and double limitation to +1 and -1
// necesarry for fls
float GainLim (float iniVal, float gain) {
  iniVal = iniVal*gain;
  if (iniVal > 1)
    iniVal = 1;
  if (iniVal < -1)
    iniVal = -1;
  return iniVal; 
}

void loop() {
  // put your main code here, to run repeatedly:
 // Serial.println("Requesting temperatures...");
  sensors.requestTemperatures(); // sent the comand to get temperatures
 // Serial.print ("The temperature is: ");
 //  Serial.println(sensors.getTempCByIndex(0));
  tempRead=sensors.getTempCByIndex(0); // can be more than one sensor in a chain
//                                   // 0 index refers to thye first one
//  Serial.println(tempRead,2);

// displaying the temperature
  LCDdisplay.setCursor(12,0); // initial position for text, 1st row
  LCDdisplay.print("Temp ");
  LCDdisplay.setCursor(9,1); // initial position for text, 2nd row
  LCDdisplay.print(tempRead);
  LCDdisplay.setCursor(14,1); // initial position for text, 2nd row
  LCDdisplay.print(" C");

// --------    process initial values of err and cerr
  err = tempSetPoint-tempRead;
  errFls = GainLim(err, gainErr);
  cerr = err - errPrev; // change of error
  errPrev = err;
  cerrFls = GainLim(cerr, gainCerr);

  // -----  display sign of temperature: >>, >, =, <, << on LCD

  if (err < -1)
    strcpy(sign, "<<");
    else if ((err > -1) && (err < -0.25))
    strcpy(sign, "< ");
    else if ((err > -0.25) && (err < 0.25))
    strcpy(sign, "= ");
    else if ((err > -.25) && (err < 1))
    strcpy(sign, "> ");
    else if (err > 1)
    strcpy(sign, ">>");
  
   LCDdisplay.setCursor(8,0); // initial position for text, 1st row
   LCDdisplay.print(sign);

// - ----  slf starts here --------------------------------------------------------

/* Zero order TS
  two inputs: error, change-of-error;  one output: incremental control
  input domain [-1; 1] for both variables output doamain [-1; 1]
  standard 9 rules;  and method - prod;  deffuzification - wtaverage
    
  Fuzzy rules table (base) with 9 rules

     err  Neg  Zero  Pos
  cerr
  Neg       N     N     Z
  Zero      N     Z     P
  Pos       Z     P     P
*/


   // membership degrees for current error to each associated fuzy sets
  for ( i = 0; i < 3; i++)  {
    mdMfErr[i] = FuzzyMf.triMf(errFls, errMf[i]); // call 'triMf' function from library FuzzyGbr
  }

// membership degrees for current change-of-error to each associated fuzy sets
  for ( i = 0; i < 3; i++)  {
    mdMfCerr[i] = FuzzyMf.triMf(cerrFls, cerrMf[i]); // call 'triMf' function from library FuzzyGbr
  }

// evaluate the firing degres of all rules in the rule base
  for (i = 0; i < 3; i++) {
    for (j = 0; j < 3; j++ )  {
      fdRule[i][j]=mdMfErr[i]*mdMfCerr[j];  // and method - prod
    }
  }

// weighted average defuzzification
  ws = 0; // initialize variable for defuzzification
  sfd = 0;  // initialize variable for defuzzification
  for (i = 0; i < 3; i++) {
    for (j = 0; j < 3; j++ )  {
      ws += fdRule[i][j]*duMf[i][j];
      sfd += fdRule[i][j];
    }
  }
  du = ws / sfd;
  du=du*gainDu; //
  // -----------------------  slf ends here ---------------------------

// ----------  compute the control signal 'u' -----------------
/* The control signal is intended to be applied to the control input of TCA785
 *  low value of control voltage - low firing angle of the SCR - high power in the load
 *  the law used to update the control signal is: u = u - du
 */

  u = u - du;
  if (u < 0)
  u = 0;
  if (u > 4095)  // maximum value for dac convertor - 12 bits
  u = 4095;
  controlVoltage = (3.3/4095)*u; // just for disply
  
  dacI2C.setVoltage(round(u), false); // send the control value (integer) through the dac

// -------- Serial monitor  -----------------------------------

  Serial.print("   ");
  Serial.print(tempSetPoint,2);
  Serial.print("  ");
  Serial.print(tempRead,2);
  Serial.print("  ");
  Serial.print(err,4);
  Serial.print("   ");
  Serial.print(cerr,4);
  Serial.print("   ");
  Serial.print(errFls,4);
  Serial.print("   ");
  Serial.print(cerrFls,4);
  Serial.print("   ");
  Serial.print(du,4);
  Serial.print("   ");
  Serial.print(u,4);
  Serial.print("   ");
  Serial.println(controlVoltage,4);
  
 // delay (10000); // pause in ms   10000ms=10s;  sample period
  delay (5000); // pause in ms   5000ms=5s;  sample period
  }
